import os

import numpy as np
import matplotlib.pyplot as plt
import pandas as pd


class WAW(object):

    '''Open .waw file.

    Args:
        file_obj (file-like object): optional open file-like object

    Attributes:
        names (list): Log names (str)
        units (list): Log units (str)
        data (numpy.ndarray): numpy array of data
        df (pandas.DataFrame): dataframe of data
        index (numpy.ndarray): reference data for the log (i.e. depth)

    '''

    def __init__(self, filename=None, file_obj=None):
        self._units = {}
        if filename is not None or file_obj is not None:
            self.read(filename=filename, file_obj=file_obj)

    def read(self, file_obj=None, filename=None):
        if not filename is None:
            file_obj = open(filename, mode='r')
        for i, line in enumerate(file_obj.readlines()):
            line = line.strip('\n')
            if i == 0:
                names = line.split(',')
            elif i == 1:
                units = [u.strip() for u in line.split(',')]
            else:
                break
        for name, unit in zip(names, units):
            self.set_unit(name, unit)
        
        file_obj.seek(0)
        data = np.loadtxt(file_obj, skiprows=2, delimiter=',')
        self.df = pd.DataFrame(data, columns=names)

        if not filename is None:
            file_obj.close()

    def set_unit(self, name, unit):
        self._units[name] = unit

    def get_unit(self, name):
        if name in self._units:
            return self._units[name]
        elif name in self.df.columns:
            self._units[name] = ""
            return self.get_unit(name)
        else:
            raise KeyError(f"Well log '{name}' not known")
    
    @classmethod
    def from_df(cls, df, units=None):
        self = cls()
        self.df = df
        assert units is None or isinstance(units, dict)
        for log_name, log_unit in units.items():
            if log_name in self.df.columns:
                self.set_unit(log_name, log_unit)
        return self

    def to_file(self, file_obj):
        file_obj.write(",".join(self.df.columns) + "\n")
        file_obj.write(",".join([self.get_unit(col) for col in self.df.columns]) + "\n")
        self.df.to_csv(file_obj, header=False)

    def to_lasio_object(self):
        from lasio import LASFile

        las = LASFile()
        las.set_data_from_df(self.df.set_index(self.df.columns[0]))
        for curve in las.curves:
            curve.unit = self.get_unit(curve.mnemonic)
        return las
    
    def to_las(self, *args, **kwargs):
        las = self.to_lasio_object()
        return las.write(*args, **kwargs)


import csv


class Comments(object):

    def __init__(self, intervals=None):
        self.intervals = intervals

    def to_wac(self, file, title='Comments'):
        if isinstance(file, str):
            file = open(file, 'w')
        table = zip(*self.intervals)
        writer = csv.writer(file, lineterminator='\n')
        writer.writerow(['TopDepth', 'BottomDepth', title])
        writer.writerow(['m', 'm', ''])
        for row in self.intervals:
            writer.writerow(row)
        file.close()


import glob
import os

from .arraydata import WAF, WAI
from .curvedata import WAW

from .wa import Comments

__version__ = '0.1'


class UnknownFormat(IOError):
    pass


def read(filename):
    name, ext = os.path.splitext(filename)
    ext = ext.lower()
    if ext == 'waf':
        return WAF(filename)
    elif ext == 'wai':
        return WAI(filename)
    elif ext == 'waw':
        return WAW(filename=filename)
    else:
        raise UnknownFormat('Cannot read in %s files' % ext)



class MultipleExport(object):
    def __init__(self, path=None):
        self.datasets = []
        if not path is None:
            self.read(path)

    def read(self, filename):
        try:
            dset = read(fn)
            self.datasets.append(dset)
        except UnknownFormat:
            pass
        self.refresh()

    def readpath(self, path):
        fns = glob.glob(os.path.join(path, '*.w??'))
        self.readfiles(fns)
                
    def readfiles(self, *fns):
        for fn in fns:
            self.read(fn)

    # def refresh(self):
    #     import pandas as pd

    #     df = pd.DataFrame()
    #     for dset in self.datasets:
    #         if isinstance(dset, WAW):
    #             self.waw_datasets.append(dset)
    #     for i, dset in enumerate(self.waw_datasets):
    #         if not index_key in indices:
    #             indices[index_key] = 1
    #         else:
    #             indices[index_key] += 1




import os

import numpy as np
import matplotlib.pyplot as plt



class WAF(object):

    def __init__(self, fn=None):
        if fn:
            self.load(fn)

    @property
    def vampl(self):
        return self.__vampl

    @vampl.setter
    def vampl(self, value):
        if value is None:
            value = np.nanmax(self.data)
        self.__vampl = value

    def load(self, fn):
        self.fn = fn
        alldata = np.loadtxt(open(fn, mode="r"), skiprows=2, delimiter=",")
        depths = alldata[:, 0]
        with open(fn, mode="r") as f:
            line = f.readline()
        times = np.array([float(s.split()[0]) for s in line.split(",")[1:]])
        data = alldata[:, 1:]
        self.generate(data, depths, times)

    def generate(self, data, depths, times, parent=None, vampl=None):
        self.data = data
        self.depths = depths
        self.times = times
        self.n = len(self.depths)
        self.m = len(self.times)

        self.cmap = plt.cm.gray
        self.vampl = vampl
        if parent:
            self.vampl = parent.vampl
            self.cmap = parent.cmap

    def plot_amplitude_hist(self, fig=None, ax=None):
        if ax is None:
            if fig is None:
                fig = plt.figure()
            ax = fig.add_subplot(111)
        amplitudes = ax.hist(self.data.ravel(), bins=30, log=True)

    def print_depths(self):
        print('%s %s %s' % ("depth range", depths[:3], "..."))
        print('%s %s %s' % (depths[self.n-3:], "n =", self.n))
        print('%s %s %s %s %s %s' % ("time range", times[:3], "...", times[self.m-3:], "m =", self.m))
        print(str(self.data.shape))

    def imshow(self, vampl=None, fig=None, figsize=None, ax=None, **kws):
        if ax is None:
            if figsize:
                fig = plt.figure(figsize=figsize)
            elif fig is None: 
                fig = plt.figure()
            ax = fig.add_subplot(111)
        if vampl:
            self.vampl = vampl
        defkws = dict(
            cmap=self.cmap, vmin=-1*self.vampl, vmax=self.vampl,
            interpolation="nearest", origin="upper",
            # left, right, bottom, top
            aspect="auto", extent=[
                self.times[0], self.times[-1],
                self.depths[-1], self.depths[0]])
        defkws.update(**kws)
        ax.imshow(self.data, **defkws)

    def extract(self, drange=None, trange=None):
        if drange is None:
            drange = (None, None)
        if trange is None:
            trange = (None, None)
        d0, d1 = drange
        if d0 is None:
            d0i = 0
        else:
            d0i = np.argmin((self.depths - d0) ** 2)
        if d1 is None:
            d1i = len(self.depths)
        else:
            d1i = np.argmin((self.depths - d1) ** 2)
        t0, t1 = trange
        if t0 is None:
            t0i = 0
        else:
            t0i = np.argmin((self.times - t0) ** 2)
        if t1 is None:
            t1i = len(self.times)
        else:
            t1i = np.argmin((self.times - t1) ** 2)
        # print('%s %s %s %s' % (d0, d1, d0i, d1i))
        # print('%s %s %s %s' % (t0, t1, t0i, t1i))
        depths = self.depths[d0i: d1i]
        times = self.times[t0i: t1i]
        new = WAF()
        new.generate(self.data[d0i: d1i, t0i: t1i], depths, times, parent=self)
        return new

    def htrace(self, depth, retdepth=False):
        di = np.argmin((self.depths - depth) ** 2)
        if retdepth:
            return self.depths[di], self.data[di, :]
        else:
            return self.data[di, :]

    def vtrace(self, time, rettime=False, interptime=True):
        ti = np.argmin((self.times - time) ** 2)
        if interptime:
            if ti < len(self.times) - 1:
                tiextra = ti + 1
            elif ti > 0:
                tiextra = ti - 1
            ti1, ti2 = sorted([ti, tiextra])
            damp = self.data[:, ti2] - self.data[:, ti1]
            dt = self.times[ti2] - self.times[ti1]
            plusamp = (damp / dt) * (time - self.times[ti1])
            return self.data[:, ti1] + plusamp
        else:
            if rettime:
                return self.times[ti], self.data[:, ti]
            else:
                return self.data[:, ti]




class WAI(object):

    def __init__(self, fn=None):
        self.__vmin = None
        self.__vmax = None
        if fn:
            self.load(fn)

    @property
    def vmax(self):
        return self.__vmax

    @vmax.setter
    def vmax(self, value):
        if value is None:
            value = np.nanmax(self.data)
        self.__vmax = value

    @property
    def vmin(self):
        return self.__vmin

    @vmin.setter
    def vmin(self, value):
        if value is None:
            value = np.nanmin(self.data)
        self.__vmin = value

    def load(self, fn):
        self.fn = fn
        alldata = np.loadtxt(open(fn, mode="r"), skiprows=2, delimiter=",")
        depths = alldata[:, 0]
        with open(fn, mode="r") as f:
            line = f.readline()
            line2 = f.readline()
        # Check for unit. Hardcoded as 0.1 us for now.
        data = alldata[:, 1:] / 10
        nsegments = data.shape[1]
        azimuths = np.linspace(0, 360, nsegments)
        self.generate(data, depths, azimuths)

    def generate(self, data, depths, azimuths, parent=None, vampl=None):
        self.data = data
        self.depths = depths
        self.azimuths = azimuths
        self.n = len(self.depths)
        self.m = len(self.azimuths)

        self.cmap = plt.cm.gray
        self.vampl = vampl
        if parent:
            self.vampl = parent.vampl
            self.cmap = parent.cmap

    def plot_amplitude_hist(self, fig=None, ax=None):
        if ax is None:
            if fig is None:
                fig = plt.figure()
            ax = fig.add_subplot(111)
        amplitudes = ax.hist(self.data.ravel(), bins=30, log=True)

    def print_depths(self):
        print('%s %s %s' % ("depth range", self.depths[:3], "..."))
        print('%s %s %s' % (self.depths[self.n-3:], "n =", self.n))
        print('%s %s %s' % ("azimuths range", self.azimuths[:3], "..."))
        print('%s %s %s' % (self.azimuths[self.m-3:], "m =", self.m))
        print(str(self.data.shape))

    def imshow(
            self, vmin=None, vmax=None, vampl=None, fig=None, ax=None, **kws):
        if ax is None:
            if fig is None:
                fig = plt.figure()
            ax = fig.add_subplot(111)
        if vampl:
            vmin = -1 * vampl
            vmax = vampl
        defkws = dict(
            cmap=self.cmap, vmin=vmin, vmax=vmax, interpolation="nearest",
            origin="upper", aspect="auto", extent=[
                self.azimuths[0], self.azimuths[-1],
                self.depths[-1], self.depths[0]])
        defkws.update(**kws)
        ax.imshow(self.data, **defkws)

    def plot_tt_slice(self, depth, fig=None, ax=None, **kws):
        if ax is None:
            if fig is None:
                fig = plt.figure()
            ax = fig.add_subplot(111, polar=True)
        ax.set_theta_direction(-1)
        ax.set_theta_offset(np.pi / 2.)
        actualdepth, data = self.htrace(depth, retdepth=True)
        thetas = np.deg2rad(self.azimuths)
        ax.plot(thetas, data, **kws)
        ax.set_rlim(self.vmin, self.vmax)
        return ax

    def extract(self, drange=None, azrange=None):
        if drange is None:
            drange = (None, None)
        if azrange is None:
            azrange = (None, None)
        d0, d1 = drange
        if d0 is None:
            d0i = 0
        else:
            d0i = np.argmin((self.depths - d0) ** 2)
        if d1 is None:
            d1i = len(self.depths) - 1
        else:
            d1i = np.argmin((self.depths - d1) ** 2)
        t0, t1 = azrange
        if t0 is None:
            t0i = 0
        else:
            t0i = np.argmin((self.azimuths - t0) ** 2)
        if t1 is None:
            t1i = len(self.azimuths) - 1
        else:
            t1i = np.argmin((self.azimuths - t1) ** 2)
        print('%s %s %s %s' % (d0, d1, d0i, d1i))
        print('%s %s %s %s' % (t0, t1, t0i, t1i))
        depths = self.depths[d0i: d1i]
        azimuths = self.azimuths[t0i: t1i]
        new = FWSWaf()
        new.generate(
            self.data[d0i: d1i, t0i: t1i], depths, azimuths, parent=self)
        return new

    def htrace(self, depth, retdepth=False):
        di = np.argmin((self.depths - depth) ** 2)
        if retdepth:
            return self.depths[di], self.data[di, :]
        else:
            return self.data[di, :]

    def vtrace(self, time, rettime=False, interptime=True):
        ti = np.argmin((self.azimuths - azimuth) ** 2)
        if interptime:
            if ti < len(self.azimuths) - 1:
                tiextra = ti + 1
            elif ti > 0:
                tiextra = ti - 1
            ti1, ti2 = sorted([ti, tiextra])
            damp = self.data[:, ti2] - self.data[:, ti1]
            dt = self.azimuths[ti2] - self.azimuths[ti1]
            plusamp = (damp / dt) * (time - self.azimuths[ti1])
            return self.data[:, ti1] + plusamp
        else:
            if rettime:
                return self.azimuths[ti], self.data[:, ti]
            else:
                return self.data[:, ti]



import glob
import sys
import os

import numpy as np
import pandas as pd
import pyqtgraph as pg
from pyqtgraph.Qt import QtCore, QtGui

from . import arraydata



class SonicViewerWindow(QtGui.QMainWindow):
    def __init__(self):
        super(SonicViewerWindow, self).__init__()
        fns = sorted(glob.glob("RX*Wide Band.waf"))
        tab_widget = QtGui.QTabWidget()
        self.setCentralWidget(tab_widget)
        for fn in fns:
            print("Loading %s..." % fn)
            log = SonicLog(fn)
            tab_widget.addTab(log, fn)
        self.showMaximized()



class SonicLog(QtGui.QWidget):
    def __init__(self, waf_fn):
        super(SonicLog, self).__init__()

        self.waf = arraydata.WAF(waf_fn)
        # self.setWindowTitle(waf_fn)
        self.waf_fn = waf_fn
        self.vdl = VDL(self.waf)
        self.time_plot = pg.PlotWidget()
        self.vertical_plot_panel = QtGui.QWidget()
        layout1 = QtGui.QVBoxLayout()
        self.vertical_plot_panel.setLayout(layout1)
        self.vertical_plot = pg.PlotWidget()
        self.export_button = QtGui.QPushButton("Export CSV")
        layout1.addWidget(self.vertical_plot)
        layout1.addWidget(self.export_button)

        self.topbottom_splitter = QtGui.QSplitter()
        self.leftright_splitter = QtGui.QSplitter()
        self.topbottom_splitter.setOrientation(QtCore.Qt.Vertical)
        self.leftright_splitter.setOrientation(QtCore.Qt.Horizontal)
        
        layout = QtGui.QHBoxLayout()
        layout.addWidget(self.leftright_splitter)
        self.setLayout(layout)
        
        self.leftright_splitter.addWidget(self.vertical_plot_panel)
        self.leftright_splitter.addWidget(self.topbottom_splitter)

        self.topbottom_splitter.addWidget(self.vdl)
        self.topbottom_splitter.addWidget(self.time_plot)
        
        
        self.rois = []
        self.add_roi(DepthSlice(self.vdl, self.time_plot))
        self.add_roi(FixedWindow(200, 50, self.vdl, self.vertical_plot, self.time_plot))
        self.export_button.clicked.connect(self.export)

        self.vdl.set_vampl(500)
        self.vdl.sigVamplChanged.connect(self.set_vampl)

    def add_roi(self, roi):
        roi.show()
        self.rois.append(roi)

    def set_vampl(self, value):
        value = abs(value) * 2
        self.time_plot.getPlotItem().setYRange(value * -1, value, padding=0)

    def export(self):
        fixed_window = self.rois[1]  # very hard-coded thingy
        t0 = fixed_window.t0
        width = fixed_window.width
        fn = self.waf_fn.replace(".waf", "_%.0f+%.0fus.csv" % (fixed_window.t0, fixed_window.width))
        depths, times, values = fixed_window.data()
        depths, times, values = resample([depths, times], [depths, values])
        keys = [
            "Depth",
            "Time (%.0f-%.0f us)" % (t0, t0 + width),
            "Amplitude (%.0f-%.0f us)" % (t0, t0 + width)
            ]
        data = {
            "Depth": depths,
            "Time (%.0f-%.0f us)" % (t0, t0 + width): times,
            "Amplitude (%.0f-%.0f us)" % (t0, t0 + width): values,
            }
        pd.DataFrame(data, columns=keys).to_csv(fn, index=False)



def resample(*pairs):
    new_pairs = []
    for pair in pairs:
        i = np.argsort(pair[0])
        p0 = pair[0][i]
        p1 = pair[1][i]
        new_pairs.append((p0, p1))
    print(new_pairs)
    index_array = np.asarray(new_pairs[0][0])
    min_gradients = [np.min(index_array)]
    for pair in new_pairs[1:]:
        index_array = np.append(index_array, pair[0])
        min_gradients.append(np.min(np.gradient(pair[0])))
    print(min_gradients)
    index_reg = np.arange(np.min(index_array), np.max(index_array), np.min(min_gradients) / 5.)
    data_regs = []
    for index_array, data in new_pairs:
        data_reg = np.interp(index_reg, index_array, data, left=np.nan, right=np.nan)
        data_regs.append(data_reg)
    return [index_reg] + data_regs




class DepthSlice(object):
    def __init__(self, vdl, time_plot, depth=None):
        self.vdl = vdl
        self.time_plot = time_plot
        if depth is None:
            depth = np.mean(vdl.waf.depths)
        self.depth = depth
        self.vdl_line = pg.InfiniteLine(depth, angle=0, movable=True, bounds=(vdl.waf.depths[0], vdl.waf.depths[-1]))
        self.trace_line = pg.PlotDataItem(self.vdl.waf.times, self.trace)
        self.vdl_line.sigPositionChanged.connect(self.update)

    @property
    def trace(self):
        return self.vdl.waf.htrace(self.depth)
        
    def update(self):
        self.depth = self.vdl_line.value()
        self.trace_line.setData(self.vdl.waf.times, self.trace)
        self.time_plot.setTitle(str(self.depth))

    def show(self):
        self.vdl.addItem(self.vdl_line)
        self.time_plot.addItem(self.trace_line)
        
    def hide(self):
        self.vdl.removeItem(self.vdl_line)
        self.time_plot.removeItem(self.trace_line)



class FixedWindow(object):
    def __init__(self, t0, width, vdl, vertical_plot, time_plot):
        self.vdl = vdl
        self.vertical_plot = vertical_plot
        self.time_plot = time_plot

        self.vdl_roi = pg.LinearRegionItem(values=[t0, t0+width], orientation=pg.LinearRegionItem.Vertical)
        self.time_plot_roi = pg.LinearRegionItem(values=[t0, t0+width], orientation=pg.LinearRegionItem.Vertical)
        depths, times, values = self.data()
        self.statistic_value_curve = pg.PlotDataItem(values, depths)
        self.statistic_time_curve = pg.PlotDataItem(times, depths, pen="r")

        self.vdl_roi.sigRegionChanged.connect(self.moved_vdl_roi)
        self.time_plot_roi.sigRegionChanged.connect(self.moved_time_plot_roi)
        self.vdl_roi.sigRegionChanged.connect(self.update)

    @property
    def t0(self):
        return min(self.vdl_roi.getRegion())

    @property
    def t1(self):
        return max(self.vdl_roi.getRegion())

    @property
    def width(self):
        return self.t1 - self.t0

    def update(self):
        depths, times, values = self.data()
        self.statistic_value_curve.setData(values, depths)
        self.statistic_time_curve.setData(times, depths)

    def data(self, statistic="min(positive)"):
        try:
            self.window_waf = self.vdl.waf.extract(
                    trange=self.vdl_roi.getRegion())
        except ValueError:
            return [], [], []
        if statistic == "max(positive)":
            seek_data = self.window_waf.data
            return_data = self.window_waf.data
            func = np.argmax
        if statistic == "min(positive)":
            seek_data = self.window_waf.data * -1
            return_data = self.window_waf.data
            func = np.argmax
        if statistic == "max(abs)":
            seek_data = np.abs(self.window_waf.data)
            return_data = np.abs(self.window_waf.data)
            func = np.argmax
        time_args = func(seek_data, axis=1)
        times = np.asarray([self.window_waf.times[ti] for ti in time_args])
        values = []
        for i in range(len(time_args)):
            values.append(return_data[i, time_args[i]])
        values = np.asarray(values)
        # print("vdl_waf = %s" % (self.vdl.waf.depths.shape, ))
        # print("window_waf = %s" % (self.window_waf.depths.shape, ))
        return self.window_waf.depths, times, values

    def moved_vdl_roi(self):
        self.time_plot_roi.setRegion(self.vdl_roi.getRegion())

    def moved_time_plot_roi(self):
        self.vdl_roi.setRegion(self.time_plot_roi.getRegion())

    def show(self):
        self.vdl.addItem(self.vdl_roi)
        self.time_plot.addItem(self.time_plot_roi)
        self.vertical_plot.addItem(self.statistic_value_curve)
        self.vertical_plot.invertY()
        self.vdl.addItem(self.statistic_time_curve)

    def hide(self):
        self.vdl.removeItem(self.vdl_roi)
        self.time_plot.removeItem(self.time_plot_roi)
        self.vertical_plot.removeItem(self.statistic_value_curve)
        self.vdl.removeItem(self.statistic_time_curve)



class VDL(pg.ImageView):

    sigVamplChanged = QtCore.Signal(float)

    def __init__(self, waf):
        super(VDL, self).__init__(view=pg.PlotItem())
        self.waf = waf
        self.view.removeItem(self.roi)
        self.view.removeItem(self.normRoi)
        self.ui.roiBtn.hide()
        hist_region = self.getHistogramWidget().item.region
        hist_region.lines[0].sigPositionChanged.connect(self.hist_drag_0_slot)
        hist_region.lines[1].sigPositionChanged.connect(self.hist_drag_1_slot)
        x0 = waf.times[0]
        x1 = waf.times[-1]
        y1 = waf.depths[-1]
        y0 = waf.depths[0]
        # print("VDL WAF data shape = %s" % (waf.data.shape, ))
        xscale = (x1 - x0) / waf.data.shape[1]
        yscale = (y1 - y0) / waf.data.shape[0]
        self.setImage(waf.data, pos=[x0, y0], scale=[xscale, yscale])
        self.view.setAspectLocked(False)

    def hist_drag_0_slot(self):
        hist_region = self.getHistogramWidget().item.region
        value = hist_region.lines[0].value()
        hist_region.lines[1].setValue(value * -1)
        self.sigVamplChanged.emit(abs(value))

    def hist_drag_1_slot(self):
        hist_region = self.getHistogramWidget().item.region
        value = hist_region.lines[1].value()
        hist_region.lines[0].setValue(value * -1)
        self.sigVamplChanged.emit(abs(value))

    def set_vampl(self, value):
        value = abs(value)
        hist = self.getHistogramWidget()
        hist.setLevels(value * -1, value)



def main():
    pg.setConfigOptions(imageAxisOrder='row-major')
    app = QtGui.QApplication([])
    app.setStyleSheet("QWidget { font-size: 0.9em; font-family: Verdana;}")
    # window = SonicViewerWindow(sys.argv[1])
    window = SonicViewerWindow()
    window.show()
    sys.exit(app.exec_())


if __name__ == '__main__':
    main()
    
